unit PenDraw_05;

// Demo:
// TFreeHandLine drawing tool for making free-hand drawing on the chart
// allows to draw arrows, lines and closed lines, single/multiline
// line selection, and deletion.
// author odissey1
// tested D2007/TC_Pro_v7.12 

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, Spin
  , TeeProcs, TeEngine, Chart, Series, TeeEdit,TeeEditPro, AppEvnts, TeCanvas, TeePenDlg, TeeEdiGrad
  , TeeFreehandLine
  , ColorGrd
  ;


type TZoomBak=record
   Xmin,Xmax,Ymin,Ymax: double;
end;

  TForm1 = class(TForm)
    Chart1: TChart;
    Series1: TLineSeries;
    Panel1: TPanel;
    lbNumPtsSimple: TLabel;
    Label7: TLabel;
    lbNumPtsOrig: TLabel;
    Label6: TLabel;
    GroupBox1: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    SpinEditPrecision: TSpinEdit;
    SpinEditSimplePenWidth: TSpinEdit;
    ColorDialog1: TColorDialog;
    ChartEditor1: TChartEditor;
    ButtonEditChart: TButton;
    CheckBoxClosed: TCheckBox;
    ButtonClear: TButton;
    CheckBoxMultiline: TCheckBox;
    ButtonZoomIn: TButton;
    ButtonZoomOut: TButton;
    ButtonUndoZoom: TButton;
    ApplicationEvents1: TApplicationEvents;
    CheckBoxDrawBehind: TCheckBox;
    ButtonColor1: TButtonColor;
    CheckBoxEnableDraw: TCheckBox;
    CheckBoxActive: TCheckBox;
    Memo1: TMemo;
    ButtonDeleteLast: TButton;
    CheckBoxEnableSelect: TCheckBox;
    ButtonDeleteSelected: TButton;
    ButtonColor2: TButtonColor;
    ComboBoxEndHeaderStyle: TComboBox;
    ComboBoxStartHeaderStyle: TComboBox;
    Label3: TLabel;
    Label4: TLabel;
    Button1: TButton;
    ButtonAddLine: TButton;
    procedure FormCreate(Sender: TObject);
    procedure SpinEditPrecisionChange(Sender: TObject);
    procedure SpinEditSimplePenWidthChange(Sender: TObject);
    procedure Chart1DblClick(Sender: TObject);
    procedure ButtonEditChartClick(Sender: TObject);
    procedure CheckBoxClosedClick(Sender: TObject);
    procedure ButtonClearClick(Sender: TObject);
    procedure ButtonZoomInClick(Sender: TObject);
    procedure ButtonZoomOutClick(Sender: TObject);
    procedure ButtonUndoZoomClick(Sender: TObject);
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure CheckBoxDrawBehindClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure CheckBoxMultilineClick(Sender: TObject);
    procedure CheckBoxEnableDrawClick(Sender: TObject);
    procedure CheckBoxActiveClick(Sender: TObject);
    procedure ButtonDeleteLastClick(Sender: TObject);
    procedure CheckBoxEnableSelectClick(Sender: TObject);
    procedure ButtonDeleteSelectedClick(Sender: TObject);
    procedure ComboBoxEndHeaderStyleCloseUp(Sender: TObject);
    procedure ComboBoxStartHeaderStyleCloseUp(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Chart1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Chart1Click(Sender: TObject);
    procedure ButtonAddLineClick(Sender: TObject);
  private
    { Private declarations }
    ZoomBak: TZoomBak;
    LineTool: TFreehandLineTool;

    procedure WheelZoom(ZoomOut: boolean; MousePos: TPoint);
    procedure LineToolNewLine(Sender: TObject);
    procedure LineToolSelect(Sender: TObject);
    procedure LineToolSelecting(Sender:TFreehandLineTool; Line:TFreehandLine; var AllowSelect:Boolean);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;


implementation

{$R *.dfm}

procedure log(s: string);
begin
     Form1.Memo1.Lines.Add(s);
end;

procedure TForm1.FormCreate(Sender: TObject);

  procedure AddStrings(HeaderStyle: TComboBox);
  begin
     // THeaderStyle = (hsNone, hsOpenArrow, hsSharpArrow, hsSimpleArrow, hsFlatEnd, hsFlatBar, hsBar, hsOpenFin, hsArcFin );
     HeaderStyle.Items.Clear;//if any
     HeaderStyle.Items.Add('None');
     HeaderStyle.Items.Add('OpenArrow');
     HeaderStyle.Items.Add('SharpArrow');
     HeaderStyle.Items.Add('SimpleArrow');
     HeaderStyle.Items.Add('FlatEnd');
     HeaderStyle.Items.Add('FlatBar');
     HeaderStyle.Items.Add('Bar');
     HeaderStyle.Items.Add('OpenFin');
//     HeaderStyle.Items.Add('ArcFin');

     HeaderStyle.ItemIndex:=0;
  end;

begin
     Chart1.Align:=alClient;
     Series1.FillSampleValues(20);
     Chart1.AllowZoom:=False;
     Panel1.DoubleBuffered:=True;//else bevelouter leaves streaks on GroupBox1
     Panel1.FullRepaint:=False;//else blinking on GroupBox1
     lbNumPtsOrig.Caption   := 'n/a';
     lbNumPtsSimple.Caption := 'n/a';
     Memo1.Clear;

//     ButtonPen1.LinkPen(SeriesDraw.Pen);
//     ButtonColor1.LinkProperty(SeriesDraw, 'SeriesColor');
//     ButtonColor1.LinkProperty(SeriesDraw, 'PenColor');


     LineTool:= TFreehandLineTool.Create(Chart1);
     LineTool.Pen.Color:=$00A9FF80;//pastel Lime //todo: Set predefined Custom Colors???
     LineTool.Pen.Width:=SpinEditSimplePenWidth.Value;

     LineTool.OnNewLine:=LineToolNewLine;
     LineTool.OnSelect:=LineToolSelect;
//     LineTool.OnSelecting:=LineToolSelecting;

     ButtonColor1.LinkProperty(LineTool, 'SelectedColor');
     ButtonColor2.LinkProperty(LineTool.Pen, 'Color');

     //todo: D2007 has some bug: it doesn't hold strings in TComboBox, so I have todo it manually here
     AddStrings(ComboBoxStartHeaderStyle);
     AddStrings(ComboBoxEndHeaderStyle);

     TeeCustomEditColors:=TStringList.Create;
     TeeCustomEditColors.Assign(ColorDialog1.CustomColors);

end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
     TeeCustomEditColors:=nil; //else Err, do not Free else Err
//     TeeCustomEditColors.Free;//Err

     LineTool.Free;
end;

procedure TForm1.ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
begin
     ApplicationEvents1.OnIdle:=nil;
//     with Chart1 do caption:=format('x1:%f, x2:%f, y1:%f, y2:%f',[BottomAxis.Minimum, BottomAxis.Maximum, LeftAxis.Minimum, LeftAxis.Maximum]);

     with ZoomBak, Chart1 do begin
       Xmin:=BottomAxis.Minimum;
       Xmax:=BottomAxis.Maximum;
       Ymin:=LeftAxis.Minimum;
       Ymax:=LeftAxis.Maximum;
     end;
end;


procedure TForm1.SpinEditPrecisionChange(Sender: TObject);
begin
     LineTool.Precision:=(Sender as TSpinEdit).Value;
end;

procedure TForm1.CheckBoxClosedClick(Sender: TObject);
begin
    LineTool.CloseLines:=(Sender as TCheckBox).Checked;
end;
procedure TForm1.CheckBoxDrawBehindClick(Sender: TObject);
begin
     LineTool.DrawBehind:=(Sender as TCheckBox).Checked;
end;
procedure TForm1.CheckBoxMultilineClick(Sender: TObject);
begin
     LineTool.Multiline:=(Sender as TCheckBox).Checked;
end;
procedure TForm1.CheckBoxEnableDrawClick(Sender: TObject);
begin
     LineTool.EnableDraw:=(Sender as TCheckBox).Checked;
end;
procedure TForm1.CheckBoxEnableSelectClick(Sender: TObject);
begin
     LineTool.EnableSelect:=(Sender as TCheckBox).Checked;
end;
procedure TForm1.ComboBoxStartHeaderStyleCloseUp(Sender: TObject);
begin
     LineTool.StartHeaders:=THeaderStyle( (Sender as TComboBox).ItemIndex );
end;
procedure TForm1.ComboBoxEndHeaderStyleCloseUp(Sender: TObject);
begin
     LineTool.EndHeaders:=THeaderStyle( (Sender as TComboBox).ItemIndex );
end;
procedure TForm1.CheckBoxActiveClick(Sender: TObject);
begin
     LineTool.Active:=(Sender as TCheckBox).Checked;
end;
procedure TForm1.SpinEditSimplePenWidthChange(Sender: TObject);
begin
     LineTool.Pen.Width:=(Sender as TSpinEdit).Value;
end;
procedure TForm1.LineToolNewLine(Sender: TObject);
begin
     log('OnNewLine');
     ButtonDeleteLast.Enabled:=(LineTool.Lines.Count>0);
     ButtonClear.Enabled:=(LineTool.Lines.Count>0);
     ButtonDeleteSelected.Enabled:=(LineTool.SelectedCount>0);

     lbNumPtsOrig.Caption   := IntToStr(LineTool.OrigCount);
     lbNumPtsSimple.Caption := IntToStr(LineTool.SimpleCount);
end;
procedure TForm1.LineToolSelect(Sender: TObject);
begin
     log('OnSelected');
//     log(format('CelectedCount:%d',[LineTool.SelectedCount]));
     ButtonDeleteLast.Enabled:=(LineTool.Lines.Count>0);
     ButtonClear.Enabled:=(LineTool.Lines.Count>0);
     ButtonDeleteSelected.Enabled:=(LineTool.SelectedCount>0);
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
     LineTool.MoveSelectionTop;
end;

procedure TForm1.ButtonClearClick(Sender: TObject);
begin
     LineTool.Clear;
end;

procedure TForm1.ButtonAddLineClick(Sender: TObject);
var Values: array of TFloatPoint;
    XMin,XMax,Ymin,Ymax: double;
    k: integer;
begin
     Xmin:=Chart1.BottomAxis.Minimum; Xmax:=Chart1.BottomAxis.Maximum;
     Ymin:=Chart1.LeftAxis.Minimum;   Ymax:=Chart1.LeftAxis.Maximum;

     //simple 2-point chords demo->
     SetLength(Values,2);
     Values[0].X:= Xmin + random * (Xmax-Xmin);
     Values[0].Y:= Ymin + random * (Ymax-Ymin);
     Values[1].X:= Xmin + random * (Xmax-Xmin);
     Values[1].Y:= Ymin + random * (Ymax-Ymin);

     //randomize->
     LineTool.Pen.Color:= StringToColor('$'+ColorDialog1.CustomColors.ValueFromIndex[random(8)]);
     LineTool.Pen.Width:= 10+random(10);
     LineTool.StartHeaders:=THeaderStyle( 4+random(5) );
     LineTool.EndHeaders  :=THeaderStyle( 1+random(4) );

     //create line using current LineTool settings
     LineTool.CreateLine(Values);
end;

procedure TForm1.ButtonDeleteSelectedClick(Sender: TObject);
begin
     LineTool.DeleteSelectedLines;
end;

procedure TForm1.ButtonDeleteLastClick(Sender: TObject);
begin
     LineTool.DeleteLast;
end;

procedure TForm1.Chart1Click(Sender: TObject);
begin
     Chart1.SetFocus;
end;

procedure TForm1.Chart1DblClick(Sender: TObject);
begin
     ChartEditor1.Execute;
end;

procedure TForm1.Chart1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
     if (Key=VK_DELETE) then begin
       LineTool.DeleteSelectedLines;
     end;
end;

procedure TForm1.ButtonEditChartClick(Sender: TObject);
begin
     ChartEditor1.Execute;
end;

procedure TForm1.ButtonZoomInClick(Sender: TObject);
begin
     Chart1.ZoomPercent(125);//todo: 125% (?)
end;
procedure TForm1.ButtonZoomOutClick(Sender: TObject);
begin
     Chart1.ZoomPercent(50);
end;

procedure TForm1.ButtonUndoZoomClick(Sender: TObject);
begin
     with ZoomBak, Chart1 do begin
       BottomAxis.Minimum:=Xmin;
       BottomAxis.Maximum:=Xmax;
       LeftAxis.Minimum  :=Ymin;
       LeftAxis.Maximum  :=Ymax;
     end;
//     Chart1.UndoZoom;
     Chart1.BottomAxis.Automatic:=True;
     Chart1.LeftAxis.Automatic  :=True;
end;
procedure TForm1.FormMouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
     WheelZoom(True, MousePos);
     Handled:=True;
end;
procedure TForm1.FormMouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
     WheelZoom(False, MousePos);
     Handled:=True;
end;
procedure TForm1.WheelZoom(ZoomOut: boolean; MousePos: TPoint);
var Xo,Yo,V1,V2: double;
    P: TPoint;
    scale: double;
begin
     P:=Chart1.ScreenToClient(MousePos);

     if ZoomOut then scale:=2 else scale:=0.5;

     with Chart1.BottomAxis do begin
       Xo:=CalcPosPoint(P.X);
       V1:=scale * (Minimum - Xo);
       V2:=scale * (Maximum - Xo);
       SetMinMax(Xo+V1, Xo+V2);
     end;
     with Chart1.LeftAxis do begin
       Yo:=CalcPosPoint(P.Y);
       V1:=scale * (Minimum - Yo);
       V2:=scale * (Maximum - Yo);
       SetMinMax(Yo+V1, Yo+V2);
     end;
//     Chart1.Refresh;
//     Chart1.Zoomed:=True;
end;

procedure TForm1.LineToolSelecting(Sender:TFreehandLineTool; Line:TFreehandLine; var AllowSelect:Boolean);
begin
     log(format('OnSelecting: Index:%d',[Line.Index]));
end;

end.



