unit TeeFreehandLine;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, Spin
  , TeeProcs, TeEngine, Chart, Series, TeeEdit,TeeEditPro, AppEvnts, TeCanvas, TeePenDlg, TeeEdiGrad
  , math
  ;

type
  TFloat = Double;
  TFloatMatrix = array[0..2, 0..2] of TFloat;     // 3x3 TFloat precision

type
  TTransformation=class(TPersistent)
  private
    Matrix: TFloatMatrix;

    procedure Clear;
    procedure Rotate(Alpha: TFloat);
    procedure Translate(dX,dY: TFloat);
    procedure Scale(Sx, Sy: TFloat);
    function Transform(P: TPoint): TPoint; overload;
    function Transform(Poly: array of TPoint): TPointArray; overload;
    function Mult(const M1, M2: TFloatMatrix): TFloatMatrix;
  public
//    constructor Create; virtual;
  published
  end;

  TFreehandLineTool=class;

  THeaderStyle = (hsNone, hsOpenArrow, hsSharpArrow, hsSimpleArrow, hsFlatEnd, hsFlatBar, hsBar, hsOpenFin, hsArcFin );

  TFreehandLine=class(TCollectionItem)
  private
    FPen: TChartPen;              //line pen
    FCloseLine : boolean;         //draw polygons
    FStartHeader: THeaderStyle;   //start arrow available
    FEndHeader  : THeaderStyle;   //end arrow available

    FSelected  : boolean;         //line selected
    Npoints: integer;             //num of points in Values array
//    Values: array of TPointFloat;
    Values: array of TFloatPoint; //array of physical points representing a line

    procedure PrepareTransformation(P0,P1: TPoint; HeaderStyle: THeaderStyle);//P0,P1-arrow direction points
    Function GetParent: TFreehandLineTool;//(Ref.XXX)
  public
    Constructor Create(Collection:TCollection); override;
    Destructor Destroy; override;

    procedure DrawLine;
    function Clicked(X,Y:Integer; PixelsTolerance:Integer=0):Boolean;

    property Parent: TFreehandLineTool read GetParent;
    property Selected: boolean read FSelected write FSelected default False;
  published
  end;

  //(Ref.XXX) this whole mess is only to be able set Chart1 as TFreehandLine.Parent!!!
  TFreehandLines=class(TOwnedCollection)//need TOwnedCollection just to set TFreehandLine.Parent = Chart1
//  TFreehandLines=class(TCollection)
  private
  public
  published
  end;

  TFreehandLineSelecting=procedure(Sender:TFreehandLineTool; Line:TFreehandLine; var AllowSelect:Boolean) of object;

  TFreehandLineTool=class(TTeeCustomToolSeries)
  private
    FLines      : TFreehandLines;
    FButton     : TMouseButton;  //draw button
    FEnableDraw : Boolean;       //can draw
    FDrawBehind : boolean;       //draw before series
    FCloseLines : boolean;       //draw polygons
    FMultiline  : boolean;       //allow draw multiple lines
    FPrecision  : integer;       //pixel distance precision
    FTolerance  : integer;       //mouse pixels to move to indicate motion

    FStartHeaders: THeaderStyle; //show 'arrow' at the beginning of line
    FEndHeaders  : THeaderStyle; //show 'arrow' by the end of line
    FEnableSelect: Boolean;      //lines selectable

    IDrawing  : Boolean;         //is in drawing state
    PosX, PosY: integer;         //mouse pixel position
    FTransformation: TTransformation; //used to draw headers (arrows) at start and end of lines

    FSelectedColor: TColor;      //line color when selected
    FSelectedCount : integer;    //number of lines selected

    FOnNewLine    : TNotifyEvent;
    FOnSelect     : TNotifyEvent;
    FOnSelecting  : TFreehandLineSelecting;

    OrigList  : array of TPoint; //pixel coordinates of mouse motion (TeeChart: TPointArray)
    SimpleList: array of TPoint; //points reduced by Douglas-Peuckers to within FPrecision pixels

//    procedure DrawLine(const aLine: TFreehandLine);//use TFreehandLine.DrawLine instead
    procedure DrawOrigLine;
    procedure AddPointToCurve(X, Y: integer);
    procedure CreateSimplifiedPolyline; // Create polyline approximation using Douglas-Peuckers algorithm
    procedure ClipDrawingRegion;
    Function InternalClicked(X,Y:Integer):TFreehandLine; overload;
    Function Clicked(X,Y:Integer):TFreehandLine; overload;

    procedure SetLines(const Value: TFreehandLines);
    procedure SetDrawBehind(const Value: boolean);
    procedure SetCloseLines(const Value: boolean);
    procedure SetPrecision(const Value: integer);
    procedure SetSelectedColor(const Value: TColor);
    function GetSelectedCount: integer;
    function GetOrigCount: integer;
    function GetSimpleCount: integer;

  protected
    procedure ChartEvent(AEvent:TChartToolEvent); override;
    procedure ChartMouseEvent( AEvent: TChartMouseEvent; AButton:TMouseButton; AShift: TShiftState; X, Y: Integer); override;

  public
    Constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;

    procedure Clear;
    Procedure DeleteSelectedLines;
    procedure DeleteLast;
    procedure MoveSelectionTop;
    procedure CreateLine(aValues: array of TFloatPoint); //create line programmatically using physical (floating) values

//    Function Clicked(X,Y:Integer):TFreehandLine; overload;
    function AxisPoint(X,Y: double): TPoint; overload;          //not used..
    Function AxisPoint(Const P: TFloatPoint): TPoint; overload; //from physical (floating) value to screen pixel
    Function ScreenPoint(const P:TPoint):TFloatPoint;           //from screen pixels position to axis double value

  published
    property Active;//visible
//    property ClickTolerance:Integer read FClick write FClick default 0;
    property EnableDraw:Boolean read FEnableDraw write FEnableDraw default True;
    property Lines: TFreehandLines read FLines write SetLines;
    property Pen;
//    property Series;
    property DrawBehind: boolean read FDrawBehind write SetDrawBehind;
    property CloseLines: boolean read FCloseLines write SetCloseLines default False;
    property Multiline:  boolean read FMultiline  write FMultiline default False;
    property Precision: integer  read FPrecision  write SetPrecision;
    property EnableSelect: boolean read FEnableSelect write FEnableSelect default False;//True not working?
    property StartHeaders: THeaderStyle read FStartHeaders  write FStartHeaders default hsNone;
    property EndHeaders:   THeaderStyle read FEndHeaders    write FEndHeaders   default hsNone;
    property SelectedColor: TColor read FSelectedColor write SetSelectedColor default clNone;
    property SelectedCount: integer read GetSelectedCount;
    property OrigCount  : integer read GetOrigCount;  //num of points in OrigList array
    property SimpleCount: integer read GetSimpleCount;//num of points in SimpleList array

    // events
    property OnNewLine:TNotifyEvent read FOnNewLine write FOnNewLine;
    property OnSelect:TNotifyEvent read FOnSelect write FOnSelect;
    property OnSelecting:TFreehandLineSelecting read FOnSelecting write FOnSelecting;//obsolete?
  end;


implementation

uses DouglasPeuckers;


//======================================================================================================================
// TTransformation->
// borrowed from Graphics32 by Alex Denisov (unit GR32_Transforms)
//======================================================================================================================

const
  IdentityMatrix: TFloatMatrix = (
    (1, 0, 0),
    (0, 1, 0),
    (0, 0, 1));

procedure TTransformation.Clear;
begin
  Matrix := IdentityMatrix;
end;

procedure TTransformation.Rotate(Alpha: TFloat);
var S, C: TFloat;
    M: TFloatMatrix;
begin
  M := IdentityMatrix;
  S := Sin(Alpha); C := Cos(Alpha);
  M[0,0] := C;  M[1,0] := S;
  M[0,1] :=-S;  M[1,1] := C;
  Matrix := Mult(M, Matrix);
end;

procedure TTransformation.Translate(Dx, Dy: TFloat);
var M: TFloatMatrix;
begin
  M := IdentityMatrix;
  M[2,0] := Dx;
  M[2,1] := Dy;
  Matrix := Mult(M, Matrix);
end;

procedure TTransformation.Scale(Sx, Sy: TFloat);
var  M: TFloatMatrix;
begin
  M := IdentityMatrix;
  M[0,0] := Sx;
  M[1,1] := Sy;
  Matrix := Mult(M, Matrix);
end;

function TTransformation.Transform(P: TPoint): TPoint;
var X,Y: double;
begin
  X := P.X * Matrix[0,0] + P.Y * Matrix[1,0] + Matrix[2,0];
  Y := P.X * Matrix[0,1] + P.Y * Matrix[1,1] + Matrix[2,1];
  result := Point(round(X), round(Y));
end;

function TTransformation.Transform(Poly: array of TPoint): TPointArray;
var i: integer;
  X,Y: double;
begin
     setlength(result,length(Poly));
     for i := 0 to length(Poly) - 1 do begin
       X := Poly[i].X * Matrix[0,0] + Poly[i].Y * Matrix[1,0] + Matrix[2,0];
       Y := Poly[i].X * Matrix[0,1] + Poly[i].Y * Matrix[1,1] + Matrix[2,1];
       result[i] := Point(round(X), round(Y));
     end;
end;

function TTransformation.Mult(const M1, M2: TFloatMatrix): TFloatMatrix;
var i, j: Integer;
begin
  for i := 0 to 2 do
    for j := 0 to 2 do
      Result[i, j] :=  M1[0, j] * M2[i, 0] + M1[1, j] * M2[i, 1] +  M1[2, j] * M2[i, 2];
end;

//======================================================================================================================
//<-TTransformation
//======================================================================================================================


//======================================================================================================================
// TFreehandLine->
//======================================================================================================================

Constructor TFreehandLine.Create(Collection:TCollection);
begin
  inherited;
  FPen:=TChartPen.Create(nil);
end;
Destructor TFreehandLine.Destroy;
begin
  FPen.Free;
  inherited;
end;

type
  TOwnedCollectionAccess=class(TOwnedCollection);

function TFreehandLine.GetParent: TFreehandLineTool;
begin
//  result:=TFreehandLineTool(Collection.Owner);
  result:=TFreehandLineTool(TOwnedCollectionAccess(Collection).GetOwner);
end;

const
   //arrow points notation:
   //           (2) ------------------- y=0
   //         /  |  \      |        |
   //        /   |   \  aOffset     |
   //       /    |    \    |     aLength
   //     /    (end)    \ ----       |
   //  (1)      (4)       (3)---------- y=-aLength
   //  -w........0........+aWidth

   //THeaderStyle = (hsNone, hsOpenArrow, hsSharpArrow, hsSimpleArrow, hsFlatEnd, hsFlatBar, hsOpenFin, hsArcFin );
   //this is old HPGL style vector presentation of the polylines
   AOpenArrow:   array [0..2] of TPoint = ( (x:-100;y:-200), (x:0;y:0), (x:100;y:-200) );
   ASharpArrow:  array [0..3] of TPoint = ( (x:-100;y:-200), (x:0;y:0), (x:100;y:-200), (x:0;y:-160) );
   ASimpleArrow: array [0..2] of TPoint = ( (x:-100;y:-200), (x:0;y:0), (x:100;y:-200) );
//   AFlatBar:     array [0..3] of TPoint = ( (x:-100;y:0), (x:-100;y:50), (x:100;y:50), (x:100;y:0) );//T-shape flat line termination
   AFlatBar:     array [0..5] of TPoint = ( (x:0;y:0), (x:-100;y:0), (x:-100;y:50), (x:100;y:50), (x:100;y:0), (x:0;y:0) );//T-shape flat line termination
   AFlatEnd:     array [0..3] of TPoint = ( (x:-50;y:0),  (x:-50;y:50),  (x:50;y:50),  (x:50;y:0)  );//flat line termination
   ABar:         array [0..1] of TPoint = ( (x:-100;y:0), (x:100;y:0) );                             //T-shape round line termination
//   AOpenFin:     array [0..2] of TPoint = ( (x:-100;y:50), (x:0;y:0), (x:100;y:50) );              //simple arrow fins
   AOpenFin:     array [0..9] of TPoint = ( (x:0;y:0),  (x:-100;y:50), (x:0;y:0),  (x:100;y:50), (x:0;y:0),
                                            (x:0;y:120),(x:-100;y:170),(x:0;y:120),(x:100;y:170),(x:0;y:120) );  //Double arrow fins
//   AOpenFin:     array [0..14] of TPoint = ( (x:0;y:-120),(x:-80; y:-70),(x:0;y:-120),(x:80;y:-70),(x:0;y:-120),
//                                             (x:0;y:0),   (x:-90;y:50),  (x:0;y:0),   (x:90;y:50), (x:0;y:0),
//                                             (x:0;y:120), (x:-100;y:170),(x:0;y:120), (x:100;y:170),(x:0;y:120) );//Triple arrow fins

procedure TFreehandLine.PrepareTransformation(P0,P1: TPoint; HeaderStyle: THeaderStyle);//PO-from point, P1-to point
var alpha: double;
    aWidth,aLength,aOffset: integer;//arrow Width, Length, Offset
begin
    alpha:=arctan2(P0.X - P1.X, P0.Y - P1.Y);//arrow direction angle //todo: doesn't work well for low zoom

    //use automatic header size proportional to pen width->
    if (HeaderStyle = hsFlatEnd)  then begin
      aWidth :=FPen.Width;
      aLength:=FPen.Width;
    end
    else
    if (HeaderStyle = hsFlatBar)  then begin
      aWidth :=max(12, FPen.Width);
      aLength:=max(4, FPen.Width);
    end
    else
    if (HeaderStyle = hsOpenFin) then begin
      aWidth :=max(12, FPen.Width);
      aLength:=max(6, FPen.Width);
    end
    else begin
      aWidth :=max(12, FPen.Width); //preferred arrow width
      aLength:=max(12, FPen.Width); //preferred arrow length
    end;

    aOffset:=0;                //open arrow
    if (HeaderStyle = hsSharpArrow) or (HeaderStyle = hsSimpleArrow) then
      aOffset:=round(FPen.Width * sqrt(5/4));//triangle,  FPen.Width * sqrt(5/4) is min offset to cover round ends

    Parent.FTransformation.Clear;
    Parent.FTransformation.Scale(aWidth/100, aLength/100); //100 is 100% scale used in const array e.g.: ASharpArrow
    Parent.FTransformation.Translate(0, aOffset);          //extra offset to cover pen radius
    Parent.FTransformation.Rotate(alpha);
    Parent.FTransformation.Translate(P0.X, P0.Y);          //move to the first point in polyline
end;

function TFreehandLine.Clicked(X,Y:Integer; PixelsTolerance:Integer=0):Boolean;
var i: integer;
    P, tmp1,tmp2: TPoint;

  function PointInPolyline(P: TPoint; PolyArrow: TPointArray; PixelsTolerance: integer): boolean;
  var i: integer;
  begin
    for i:=high(PolyArrow) downto 1 do begin
      result:=PointInLine(P,PolyArrow[i],PolyArrow[i-1],PixelsTolerance);
      if result then break;
    end;
  end;

  function ClickPolyLine(Poly: array of TPoint): boolean;
  var PolyArrow: TPointArray;
  begin
    PolyArrow := Parent.FTransformation.Transform(Poly);
    result := PointInPolyline(P, PolyArrow, PixelsTolerance);
  end;

  function ClickPolyGon(Poly: array of TPoint): boolean;
  var PolyArrow: TPointArray;
  begin
    PolyArrow := Parent.FTransformation.Transform(Poly);
    result := PointInPolygon(P, PolyArrow);
  end;

  function HeaderClicked(P0,P1: TPoint; HeaderStyle: THeaderStyle): boolean;
  begin
    PrepareTransformation(P0,P1, HeaderStyle);

    result:=False;
    if HeaderStyle = hsOpenArrow   then result:=ClickPolyLine(AOpenArrow)  else
    if HeaderStyle = hsBar         then result:=ClickPolyLine(ABar)        else
    if HeaderStyle = hsOpenFin     then result:=ClickPolyLine(AOpenFin)    else
    if HeaderStyle = hsSharpArrow  then result:=ClickPolyGon(ASharpArrow)  else
    if HeaderStyle = hsSimpleArrow then result:=ClickPolyGon(ASimpleArrow) else
    if HeaderStyle = hsFlatEnd     then result:=ClickPolyGon(AFlatEnd)     else
    if HeaderStyle = hsFlatBar     then result:=ClickPolyGon(AFlatBar);
  end;

begin
     P:=TeePoint(X,Y);//hit point

     PixelsTolerance:=max((FPen.Width div 2) - 1, 0);

     if FStartHeader <> hsNone then begin
       //todo: this can result in tmp1=tmp2  and lost direction at low zoom
       tmp1 :=Parent.AxisPoint(Values[0]);               //FromPoint
       tmp2 :=Parent.AxisPoint(Values[1]);               //ToPoint
       result := HeaderClicked(tmp1, tmp2, FStartHeader);//t1-FromPoint, t2-ToPoint;
       if result then exit;                              //clicked
     end;

     if FEndHeader <> hsNone then begin
       tmp1 :=Parent.AxisPoint(Values[Npoints-1]);       //FromPoint
       tmp2 :=Parent.AxisPoint(Values[Npoints-2]);       //ToPoint
       result := HeaderClicked(tmp1, tmp2, FEndHeader);  //t1-FromPoint, t2-ToPoint;
       if result then exit;                              //clicked
     end;

     //is line clicked?->
     i:=Npoints;
     dec(i); tmp2:=Parent.AxisPoint(Values[i]);
     repeat //in reverse order
       tmp1:=tmp2;
       dec(i); tmp2:=Parent.AxisPoint(Values[i]);
       result:=PointInLine(P,tmp1,tmp2,PixelsTolerance);
       if result then exit;//clicked
     until (i=0);

     //check closing chord->
     if FCloseLine then begin
       tmp2:=Parent.AxisPoint(Values[Npoints-1]);
       tmp1:=Parent.AxisPoint(Values[0]);
       result:=PointInLine(P,tmp1,tmp2,PixelsTolerance);
       if result then exit;//clicked
     end;
end;


procedure TFreehandLine.DrawLine;
var i: integer;
    PolyPoints: array of TPoint;

  procedure DrawPolyLine(Poly: array of TPoint);
  var PolyArrow: TPointArray;
  begin
    with Parent, ParentChart.Canvas do begin
      Pen.Width:=FPen.Width;
      PolyArrow := FTransformation.Transform(Poly);
      PolyLine(PolyArrow);
    end;
  end;

  procedure DrawPolyGon(Poly: array of TPoint);
  var PolyArrow: TPointArray;
  begin
    with Parent, ParentChart.Canvas do begin
      Pen.Width:=1;
      Brush.Style:=bsSolid;//filled
      Brush.Color:=Pen.Color;//same color
      PolyArrow := FTransformation.Transform(Poly);
      PolyGon(PolyArrow);
    end;
  end;

  procedure DrawHeader(P0,P1: TPoint; HeaderStyle: THeaderStyle );//P0-start point, P1-next point (else P0-end point, P1-previous point)
  begin
    PrepareTransformation(P0,P1, HeaderStyle);

    if HeaderStyle = hsOpenArrow   then DrawPolyLine(AOpenArrow)  else
    if HeaderStyle = hsBar         then DrawPolyLine(ABar)        else
    if HeaderStyle = hsOpenFin     then DrawPolyLine(AOpenFin)    else
    if HeaderStyle = hsSharpArrow  then DrawPolyGon(ASharpArrow)  else
    if HeaderStyle = hsSimpleArrow then DrawPolyGon(ASimpleArrow) else
    if HeaderStyle = hsFlatEnd     then DrawPolyGon(AFlatEnd)     else
    if HeaderStyle = hsFlatBar     then DrawPolyGon(AFlatBar);
  end;

begin
     //draw an individual line->

     SetLength(PolyPoints,Npoints);
     for i:=0 to Npoints-1 do
       PolyPoints[i]:=Parent.AxisPoint( Values[i] );

     with Parent, ParentChart.Canvas do begin
       AssignVisiblePen(FPen);
       if FSelected then
         ParentChart.Canvas.Pen.Color:=FSelectedColor;//use pre-defined selection color

//         AssignBrush();//??
       Brush.Style:=bsClear;

       if FCloseLine
       then PolyGon(PolyPoints)
       else PolyLine(PolyPoints);
     end;

     //draw headers->
     if FStartHeader <> hsNone then DrawHeader(PolyPoints[0],PolyPoints[1], FStartHeader);              //Draw StartHeader
     if FEndHeader <> hsNone   then DrawHeader(PolyPoints[Npoints-1],PolyPoints[Npoints-2], FEndHeader);//Draw EndHeader
end;

//======================================================================================================================
//<-TFreehandLine
//======================================================================================================================



//======================================================================================================================
// TFreehandLineTool->
//======================================================================================================================

Constructor TFreehandLineTool.Create(AOwner: TComponent);
begin
  inherited;
  Self.ParentChart:=TChart(AOwner);
//  FLines:=TFreehandLines.Create(TFreehandLine);//TCollection
  FLines:=TFreehandLines.Create(Self,TFreehandLine);//TOwnedCollection

  FButton := mbLeft;
  FEnableDraw:=True;  //default=True not working?
  FEnableSelect:=True;//default=True not working?
  FDrawBehind:=True;
  FPrecision:=2;
  FTolerance:=3;

  Pen.Color := $00A9FF80;//pastel Lime
  Pen.Width := 20;
  Pen.Style := psSolid;

  FTransformation:=TTransformation.Create;
end;

Destructor TFreehandLineTool.Destroy;
begin
  FLines.Free;
  FTransformation.Free;
  inherited;
end;


procedure TFreehandLineTool.Clear;
begin
     SetLength(OrigList, 0);// Delete the old curves
     SetLength(SimpleList, 0);
     FLines.Clear;
     Repaint;//Invalidate Parent Chart
     if Assigned(FOnNewLine) then FOnNewLine(Self);//todo: visible changed
end;

procedure TFreehandLineTool.ChartEvent(AEvent: TChartToolEvent);
var i: integer;
begin
  inherited;

  if ((AEvent=cteBeforeDrawSeries) and FDrawBehind)
  or ((AEvent=cteAfterDraw) and (not FDrawBehind))
  then begin

    if (Lines.Count>0) then begin //draw previous lines
      ParentChart.Canvas.BackMode:=cbmTransparent;//?
      ClipDrawingRegion;

      for i:=0 to Lines.Count-1 do
//        DrawLine(TFreehandLine(Lines.Items[i]));
        TFreehandLine(Lines.Items[i]).DrawLine;

      ParentChart.Canvas.UnClipRectangle;
    end;

    if IDrawing then
      DrawOrigLine; //Draw original (non-reduced) polyline

  end;
end;

{
//obsolete, use TFreehandLine.DrawLine instead
procedure TFreehandLineTool.DrawLine(const aLine: TFreehandLine);
var i: integer;
    PolyPoints: array of TPoint;

  procedure DrawHeader(const aLine: TFreehandLine; P0,P1: TPoint; HeaderStyle: THeaderStyle );//P0-start point, P1-next point (else P0-end point, P1-previous point)
  var alpha: double;
      aWidth,aLength,aOffset: integer;//arrow Width, Length, Offset
      PolyArrow: TPointArray;
  begin
    alpha:=arctan2(P0.X - P1.X, P0.Y - P1.Y);//arrow direction angle //todo: doesn't work well for low zoom

    if (HeaderStyle = hsFlatEnd)  then begin
      aWidth :=aLine.FPen.Width;
      aLength:=aLine.FPen.Width;
    end
    else
    if (HeaderStyle = hsFlatBar)  then begin
      aWidth :=max(12, aLine.FPen.Width);
      aLength:=aLine.FPen.Width;
    end
    else
    if (HeaderStyle = hsOpenFin) then begin
      aWidth :=max(12, aLine.FPen.Width);
      aLength:=max(6, aLine.FPen.Width);
    end
    else begin
      aWidth :=max(12, aLine.FPen.Width); //preferred arrow width
      aLength:=max(12, aLine.FPen.Width); //preferred arrow length
    end;

    aOffset:=0;                //open arrow
    if (HeaderStyle = hsSharpArrow) or (HeaderStyle = hsSimpleArrow) then
      aOffset:=round(aLine.FPen.Width * sqrt(5/4));//triangle,  FPen.Width * sqrt(5/4) is min offset to cover round ends

    FTransformation.Clear;
    FTransformation.Scale(aWidth/100, aLength/100); //100 is scale used in const array e.g.: ASharpArrow
    FTransformation.Translate(0, aOffset);//offset to cover round ends
    FTransformation.Rotate(alpha);
    FTransformation.Translate(P0.X, P0.Y);//move to the first point in polyline

    //THeaderStyle = (hsNone, hsOpenArrow, hsSharpArrow, hsTriangleArrow, hsFlatEnd, hsFlatBar, hsOpenFin, hsArcFin );

    with ParentChart.Canvas do
    if (HeaderStyle = hsOpenArrow) then begin
      PolyArrow := FTransformation.Transform(AOpenArrow);
      Pen.Style:=psSolid;
      PolyLine(PolyArrow);//open arrow
    end
    else
    if (HeaderStyle = hsBar) then begin
      PolyArrow := FTransformation.Transform(ABar);
      Pen.Style:=psSolid;
      PolyLine(PolyArrow);
    end
    else
    if (HeaderStyle = hsOpenFin) then begin
      PolyArrow := FTransformation.Transform(AOpenFin);
      Pen.Style:=psSolid;
      PolyLine(PolyArrow);
    end
    else
    if (HeaderStyle = hsSharpArrow) then begin
      PolyArrow := FTransformation.Transform(ASharpArrow);
      Pen.Style:=psClear;//sharp edges
      Brush.Style:=bsSolid;//filled
      Brush.Color:=Pen.Color;//same color
      PolyGon(PolyArrow);//closed triangle
    end
    else
    if (HeaderStyle = hsSimpleArrow) then begin
      PolyArrow := FTransformation.Transform(ASimpleArrow);
      Pen.Style:=psClear;//sharp edges
      Brush.Style:=bsSolid;//filled
      Brush.Color:=Pen.Color;//same color
      PolyGon(PolyArrow);//closed triangle
    end
    else
    if HeaderStyle = hsFlatEnd then begin
      PolyArrow := FTransformation.Transform(AFlatEnd);
      Pen.Style:=psClear;//sharp edges
      Brush.Style:=bsSolid;//filled
      Brush.Color:=Pen.Color;//same color
      PolyGon(PolyArrow);//closed triangle
    end
    else
    if HeaderStyle = hsFlatBar then begin
      PolyArrow := FTransformation.Transform(AFlatBar);
      Pen.Style:=psClear;//sharp edges
      Brush.Style:=bsSolid;//filled
      Brush.Color:=Pen.Color;//same color
      PolyGon(PolyArrow);//closed triangle
    end;

  end;


begin
//     log(format('%d, clr:%s',[index, ColorToString(aLine.FPen.Color)]));

     //draw an individual line->
     with aLine do begin

       SetLength(PolyPoints,Npoints);
       for i:=0 to Npoints-1 do
         PolyPoints[i]:=AxisPoint( Values[i] );

       with ParentChart.Canvas do begin
         AssignVisiblePen(FPen);
         if FSelected then
           ParentChart.Canvas.Pen.Color:=FSelectedColor;//use pre-defined selection color

//         AssignBrush();//??
         Brush.Style:=bsClear;

         if FCloseLine
         then PolyGon(PolyPoints)
         else PolyLine(PolyPoints);
       end;

       if FStartHeader <> hsNone then DrawHeader(aLine, PolyPoints[0],PolyPoints[1], FStartHeader);//Draw StartArrow
       if FEndHeader <> hsNone   then DrawHeader(aLine, PolyPoints[Npoints-1],PolyPoints[Npoints-2], FEndHeader);//Draw EndArrow
     end;//aLine
end;
}
procedure TFreehandLineTool.DrawOrigLine;
begin
     // Draw original (non-reduced) polyline->
     if (Length(OrigList) > 0) then
     with ParentChart.Canvas do begin

       AssignVisiblePen(Self.Pen);

       Brush.Style:=bsClear;

       if FCloseLines
       then PolyGon(OrigList)
       else PolyLine(OrigList);
    end;
end;

procedure TFreehandLineTool.ClipDrawingRegion;
var R: TRect;
begin
     if ParentChart.CanClip then begin
       with ParentChart.ChartRect do R:=Rect(Left+1,Top+1,Right,Bottom-1);//todo: ChartRect doesn't clips right
       ParentChart.Canvas.ClipRectangle(R);
     end;
end;

Procedure TFreehandLineTool.ChartMouseEvent( AEvent: TChartMouseEvent; AButton:TMouseButton; AShift: TShiftState; X, Y: Integer);

  function UnselectAll: boolean;
  var i: integer;
  begin
    result:=False;
    for i:=0 to Lines.Count - 1 do
    if TFreehandLine(Lines.Items[i]).Selected then begin
      TFreehandLine(Lines.Items[i]).Selected:=False;
      result:=True;
    end;
  end;

var
   Dist: double;
   tmpLine: TFreehandLine;
   tmpAllow: Boolean;

begin
  Case AEvent of
    cmeDown: if AButton=FButton then
             begin
               if FEnableSelect then tmpLine:=Clicked(X,Y)
                                else tmpLine:=nil;

               if Assigned(tmpLine) then // clicked line, do select
               begin

                 if Assigned(FOnSelecting) then FOnSelecting(Self,tmpLine,tmpAllow);

                 if (ssCtrl in AShift) then begin //multiple select event
                   tmpLine.Selected:=not tmpLine.Selected;//invert selecion
                   Repaint;
                 end
                 else begin //single-line selection event
                  if not tmpLine.Selected or (SelectedCount>1) then begin
                     UnselectAll;//unselect all selected (if any)..
                     tmpLine.Selected:=True;
                     Repaint;
                   end;
                 end;//single-line selection event
                 if Assigned(FOnSelect) then FOnSelect(Self);
               end //Assigned(tmpLine)
               else begin//background clicked
                 if (ssCtrl in AShift) then exit; //multiple select event -> do nothing

                 if UnselectAll then begin
                   Repaint;//unselect all selected (if any)..
                   if Assigned(FOnSelect) then FOnSelect(Self);
                 end;

                 if FEnableDraw then begin //start a new line
                   IDrawing:=True;
                   SetLength(OrigList, 0);
                   SetLength(SimpleList, 0);
                   AddPointToCurve(X, Y); // Add the new startpoint
                 end;
               end;//background clicked

             end;//AButton=FButton

    cmeMove: if IDrawing then // drawing or dragging
             begin
               Dist := TeeDistance(X - PosX, Y - PosY); //sqrt(sqr(X - PosX) + sqr(Y - PosY));
               if Dist >= FTolerance then begin
                 if not FMultiline and (FLines.Count>0) then FLines.Clear; //clear old curves once started a new one
                 AddPointToCurve(X, Y);//Add points to the curve as long as the mouse is down
                 Repaint;
               end;
             end;

      cmeUp: if AButton=FButton then
             begin
               if IDrawing then  // if drawing a new line...
               begin
                 IDrawing:=False;
                 CreateSimplifiedPolyline;// Mouse is released so finalize the curve
               end;
             end;
  end;
end;

procedure TFreehandLineTool.AddPointToCurve(X, Y: integer);
begin
     PosX := X;
     PosY := Y;
     SetLength(OrigList, Length(OrigList) + 1);
     OrigList[Length(OrigList) - 1] := Point(X,Y);
end;

procedure TFreehandLineTool.CreateSimplifiedPolyline; // Create the simple polyline approximation
var ALength: integer;
    i: integer;
    newLine: TFreehandLine;
begin

  if length(OrigList) > 1 then begin
    //Create the simple polyline approximation->
    SetLength(SimpleList, Length(OrigList));
    ALength := PolySimplifyInt2D(FPrecision, OrigList, SimpleList);
    SetLength(SimpleList, ALength);

    newLine:=TFreehandLine.Create(FLines);
    newLine.Npoints:=ALength;
    SetLength(newLine.Values, ALength);
    newLine.FCloseLine:=FCloseLines;
    if FCloseLines then newLine.FStartHeader:=hsNone else newLine.FStartHeader:=FStartHeaders;
    if FCloseLines then newLine.FEndHeader  :=hsNone else newLine.FEndHeader  :=FEndHeaders;

    newLine.FPen.Assign(Self.Pen);

    for i := 0 to ALength - 1 do
      newLine.Values[i]:=ScreenPoint( SimpleList[i] ); //from pixel to physical (floating) values

    Repaint;

    if Assigned(FOnNewLine) then FOnNewLine(Self);
  end
  else begin
    SetLength(OrigList, 0);// Delete the old curves
    SetLength(SimpleList, 0);
  end;

end;

procedure TFreehandLineTool.CreateLine(aValues: array of TFloatPoint);//allows to create lines programmatically
var ALength: integer;
    newLine: TFreehandLine;
    i: integer;
begin
    if length(aValues) = 0 then exit;
    if not FMultiline and (FLines.Count>0) then FLines.Clear; //clear old curves once started a new one

    newLine:=TFreehandLine.Create(FLines);
    newLine.FCloseLine:=FCloseLines;
    if FCloseLines then newLine.FStartHeader:=hsNone else newLine.FStartHeader:=FStartHeaders;
    if FCloseLines then newLine.FEndHeader  :=hsNone else newLine.FEndHeader  :=FEndHeaders;

    newLine.FPen.Assign(Self.Pen);

    ALength:=length(aValues);
    newLine.Npoints:=ALength;
    SetLength(newLine.Values, ALength);
    for i := 0 to ALength - 1 do
      newLine.Values[i]:=aValues[i];

    Repaint;

    if Assigned(FOnNewLine) then FOnNewLine(Self);
end;


Function TFreehandLineTool.Clicked(X,Y:Integer): TFreehandLine;
begin
     result:=InternalClicked(X,Y);
end;

Function TFreehandLineTool.InternalClicked(X,Y:Integer):TFreehandLine;
var i: Integer;
begin
  result:=nil;

  //find which line was clicked
  for i:=Lines.Count-1 downto 0 do begin //Check top lines first
    if TFreehandLine(Lines.Items[i]).Clicked(X,Y) then begin
      result:=TFreehandLine(Lines.Items[i]);
      break;
    end;
  end;
end;

function TFreehandLineTool.AxisPoint(const P: TFloatPoint): TPoint;
begin
  // convert from axis double XY to screen pixels XY
  result.X:=GetHorizAxis.CalcPosValue(P.X);
  result.Y:=GetVertAxis.CalcPosValue(P.Y);
end;
function TFreehandLineTool.AxisPoint(X,Y: double): TPoint;
begin
  // convert from axis double XY to screen pixels XY
  result.X:=GetHorizAxis.CalcPosValue(X);
  result.Y:=GetVertAxis.CalcPosValue(Y);
end;

function TFreehandLineTool.ScreenPoint(const P: TPoint): TFloatPoint;
begin
  // convert from screen pixels XY position to axis double XY
  result.X:=GetHorizAxis.CalcPosPoint(P.X);
  result.Y:=GetVertAxis.CalcPosPoint(P.Y);
end;


procedure TFreehandLineTool.SetLines(const Value: TFreehandLines);
begin
  FLines.Assign(Value);
end;
procedure TFreehandLineTool.SetDrawBehind(Const Value: boolean);
begin
  FDrawBehind:=Value;
  Repaint;
end;
procedure TFreehandLineTool.SetCloseLines(Const Value: boolean);
begin
  FCloseLines:=Value;
  Repaint;
end;
procedure TFreehandLineTool.SetPrecision(Const Value: integer);
begin
  FPrecision:=Value;
  if not FMultiline then begin
    // We must re-do the approximation
//    CreateSimplifiedPolyline;//not working..
    Repaint;
  end;
end;

procedure TFreehandLineTool.DeleteLast;
begin
     if (Lines.Count>1) then //is FMultiline
     begin
       TFreehandLine( Lines.GetItem(Lines.Count-1) ).Free; //delete last line
       SetLength(OrigList, 0);// Delete the old curves
       SetLength(SimpleList, 0);
       Repaint;
       if Assigned(FOnNewLine) then FOnNewLine(Self);
     end
     else
       Clear;     //delete all lines

end;
procedure TFreehandLineTool.DeleteSelectedLines;
var i: integer;
    need_repaint: boolean;
begin
  need_repaint:=False;
  for i:=Lines.Count-1 downto 0 do
   if TFreehandLine(Lines.Items[i]).Selected then begin
      TFreehandLine(Lines.Items[i]).Free;
      need_repaint:=True;
    end;

   if need_repaint then Repaint;
   if Assigned(FOnNewLine) then FOnNewLine(Self);

end;

procedure TFreehandLineTool.MoveSelectionTop;
begin
    //not implemented
end;

function TFreehandLineTool.GetSelectedCount: integer;
var i: integer;
begin
     result:=0;
     for i:=0 to Lines.Count-1 do
       if TFreehandLine(Lines.Items[i]).Selected then
         inc(result);
end;

procedure TFreehandLineTool.SetSelectedColor(const Value: TColor);
begin
     if Value <> FSelectedColor then begin
       FSelectedColor:=Value;
       if SelectedCount > 0 then
         Repaint;
     end;
end;

function TFreehandLineTool.GetOrigCount: integer;
begin
     result:=Length(OrigList);
end;
function TFreehandLineTool.GetSimpleCount: integer;
begin
     result:=Length(SimpleList);
end;

//======================================================================================================================
// <-TFreehandLineTool
//======================================================================================================================

end.




(*
ciuly:

aux:TCollectionItem;
aux:= FTest.items[index];
FTest.items[index]=FTest.items[index+1];
FTest.items[index+1]:=aux;

and that's about it
//==================================

ciuly:
yes . from ColnEdit:

procedure TCollectionEditor.MoveUpClick(Sender: TObject);
var
  I, InsPos: Integer;
begin
  if (ListView1.SelCount = 0) or
    (ListView1.SelCount = Collection.Count) then Exit;

  InsPos := 0;
  while not ListView1.Items[InsPos].Selected do
    Inc(InsPos);
  if InsPos > 0 then Dec(InsPos);

  Collection.BeginUpdate;
  try
     for I := 0 to ListView1.Items.Count - 1 do
       if ListView1.Items[I].Selected then
       begin
         Collection.Items[I].Index := InsPos;
         Inc(InsPos);
       end;
  finally
    Collection.EndUpdate;
  end;
  GetSelection;
  Designer.Modified;
end;

procedure TCollectionEditor.MoveDownClick(Sender: TObject);
var
  I, InsPos: Integer;
begin
  if (ListView1.SelCount = 0) or
    (ListView1.SelCount = Collection.Count) then Exit;

  InsPos := ListView1.Items.Count - 1;
  while not ListView1.Items[InsPos].Selected do
    Dec(InsPos);
  if InsPos < (ListView1.Items.Count -1) then Inc(InsPos);

  Collection.BeginUpdate;
  try
     for I := ListView1.Items.Count - 1 downto 0 do
       if ListView1.Items[I].Selected then
       begin
         Collection.Items[I].Index := InsPos;
         Dec(InsPos);
       end;
  finally
    Collection.EndUpdate;
  end;
  GetSelection;
  Designer.Modified;
end;



rllibby:
ginsonic,

The TCollection class does not expose its internal TList. But, the collection item does allow for setting of its index (which will call the Move(...) method in the internal TList). Eg:


// Move item down in list
if (X < Pred(Collection.Count)) then
  Collection.Items[X].Index:=Succ(X);

// Move item up in list
if (X > 0) and (X < Collection.Count) then
  Collection.Items[X].Index:=Succ(X);

Regards,
Russell
*)
